<?php if (!defined('PmWiki')) exit();

/*

 Username Authentification

 -- Copyright --

 Copyright 2004, by James McDuffie (ttq3sxl02@sneakemail.com)

 -- License --

 GNU GPL

 -- Description --

 This PmWiki plugin makes it possible to restrict access to the entire
 wiki by requring a username and password for certain actions. Individual
 pages and groups can still be password restricted in the same way that
 PmWiki does by default. However, user authentification happens before
 wiki password authentification. Additionally, the username doubles as
 the author name.

 By default this plugin reads username, password and ability information
 from a .htpasswd file. This file is created when the script is first run
 in the PmWiki "local" directory with the default restrictions that non 
 authorized users can only read pages. The htpasswd file has an additional
 third field where abilities are stored. This field should not interfer if
 you wish to use the same htpassword file with Apache basic authentification.
 A typical htpassword file would look like the following:

   GuestUser::read
   FredFlintstone:HWxwafxAweaXefafdae:read_all,edit
   Admin:XwasgeGGGaaceexeadeE:admin

 The abilities field is a list of actions the user can perform separated by
 commas. Abilities correspond to the same type of actions defined by PmWiki in 
 the $DefaultPassword array as well as some custom actions:

   PmWiki Actions:
     admin, attr, edit, read
   UserAuth Actions:
     admin
       - Admin is also checked explicitly by this plugin. This allows
         an admin user to be able to access any PmWiki password
         protected page without supplying the password.
     read_all  
       - Allows the user to read any page, even if it is read password
         protected.
     read_group-GroupName
       - Allows the user to read all pages in GroupName, for example:
         read_group-PrivateArea
   
 -- Installation Instructions --

 * Unpack the distribution file into your PmWiki local directory

 * Include the plugin from your config.php file:

   require_once("local/userauth.php");

 * Load your PmWiki once to create the default .htpasswd file

 * Modify the created .htpasswd file to include a user that has "edit" ability.
   Use your wiki to create the password by appending ?action=crypt to the url
   to load the PmWiki password encryption page. Create the user's line in the
   htpasswd file in the usual Apace htpasswd format but make sure to add and
   extra field to contain the user's abilites. See the description above for
   an example of how to format the file.

 * Create the page Main.LoginPage so that the login page will show up within
   your wiki layout, otherwise a default standalone page will be used. Simply
   put the following on Main.LoginPage:

     [=[loginform]=]

   Main.LoginPage is just the default and the login page can be changed to
   whatever you like. See the configuration details below. 

 -- History --

 0.1 - Initial version
 0.2 - Added ability for allowing a user to view an entire view password 
       protected group
 0.3 - Updated for PmWiki v2. (BenWilson)

 -- Configuration --

 The following variables can be defined before userauth.php is included in
 in config.php:

 $GuestUsername - Specifies the username that is used by default when no
 user is logged in.
   
 $LoginPage - Name of the page which has the "[[loginform]]" double bracket
 code on it to display the login form.

 $LackProperAbilitiesFmt - Message displayed when the user unsuccesfully
 tries to login.

 $AuthPageTitle - Title of the standalone page showing the login form.

 $HtPasswdFile - Alternative location of the HtPassword file. Not used if
 $UserInfoObj is set before the script is included in config.php.

 HtPasswd.php will be used to grab user information unless the
 $UserInfoObj is already set. This allows someone to write a new user
 info object that gets its information from a database or some other
 format. This new class would only need to implement the same minimum
 contract that HtPasswd.php does

 The HTML that is displayed on the LoginPage can be easily changed. By default
 the [loginform] inline replacement calls a function that returns the login
 page. To specify different HTML then either redefine the definition of 
 [loginform] in the Markup() or use some other method that
 correctly calls the login action.

 If you wish to change your template to include a link that displays "Login" or
 "Logout" depending on whether or not a user is logged in then add the following
 to the template:

   <a href='$PageUrl?action=$LoginAction'>$LoginName</a>

*/

define(USER_AUTH_VERSION, '0.2'); 

// PHP session related
session_start();
if (@$_REQUEST['username']) $_SESSION['username'] = $_REQUEST['username'];
if (@$_REQUEST['user_pw'])  $_SESSION['user_pw']  = $_REQUEST['user_pw'];
if (@$_REQUEST['wiki_pw'])  $_SESSION['wiki_pw']  = $_REQUEST['wiki_pw'];

// Override PmWiki variables
$Author = $_SESSION['username'];
$AuthFunction = "UserAuth";

// Configuration that your probably do not want to change
SDV($HandleActions['logout'], 'HandleLogout');
SDV($HandleActions['login'], 'HandleLogin');

SDV($LoginAction, GetLoginAction());
SDV($LoginName, ucfirst(GetLoginAction()));

// Customizable configuration variable
SDV($GuestUsername, "GuestUser");

SDV($LoginPage, "Main.LoginPage");

SDV($LackProperAbilitiesFmt, "Insufficient privileges to perform action.");

Markup("loginform", "inline","/\\[loginform\\]/", GetUserLoginForm(true));

SDV($AuthPageTitle, "Authorization required");

# HtPasswd will be used unless the $UserInfoObj is already set.
# This allows someone to write a new user info object that gets
# its information from a database or some other format. This new
# class would only need to implement the same minimum contract that
# HtPasswd.php does
if(!isset($UserInfoObj)) {
  require_once("userauth/HtPasswd.php");

  SDV($HtPasswdFile, getcwd() . "/local/.htpasswd");

  # If htpassword file does not exist then create a default empty one
  if (!file_exists($HtPasswdFile)) {
    $fp = @fopen($HtPasswdFile, "w");
    if ($fp) {
      fwrite($fp, "$GuestUsername::read\n");
      fclose($fp);
    }
  }

  // Object which contains username, password and abilities info
  $UserInfoObj = new HtPasswd($HtPasswdFile);

}


function UserAuth($pagename,$level,$authprompt=true) {
  global $GuestUsername, $HtPasswdFile, $UserInfoObj;

  $page = ReadPage($pagename);
  if (!$page) { return false; }

  // If no username has been specified then use a default name
  $auth_username = $_SESSION['username'];
  if ($auth_username=="") { $auth_username = $GuestUsername; }

  // Only return page when both user and wiki passwords work
  $user_check = 
    CheckUserPassword( $auth_username, $_SESSION['user_pw'], $level );

  $wiki_check =
    CheckWikiPassword( $_SESSION['wiki_pw'], $level, $page, $pagename );

  // To check if user can read an entire group
  $group = FmtPageName('$Group',$pagename);

  // Admin overrides everything even if a password is already specified
  if ( ($user_check && $UserInfoObj->UserHasAbility($auth_username, 'admin')) ||

       ($user_check && ($level == 'read') && 
	$UserInfoObj->UserHasAbility($auth_username, 'read_all') || 
	$UserInfoObj->UserHasAbility($auth_username, 'read_group-' . $group)
	) ||

       ($user_check && $wiki_check) ) {

    return $page;
  }

  // Display login prompt if necessary
  if (!$authprompt) return false;

  if ( !$user_check ) {
    // Remove wrong login information from session
    $_SESSION['username'] = "";
    $_SESSION['user_pw'] = "";

    HandleLogin( $pagename );
  } elseif ( !$wiki_check ) {
    // Remove password information from session
    $_SESSION['wiki_pw'] = "";
    
    PrintEmbeddedPage( $pagename, GetWikiPasswordForm($pagename) );
  }
 
  exit;
}

function CheckUserPassword($auth_username, $auth_password, $level) {
  global $GuestUsername, $LackProperAbilitiesFmt, $UserInfoObj;

  // Access username from htpasswd file and see if level and password match
  $stored_password = $UserInfoObj->GetUserPassword($auth_username);

  // Only allow an empty password for the guest user
  if ( ( $auth_username == $GuestUsername && 
	 $stored_password == "" && $auth_password == "") ||
       crypt($auth_password, $stored_password) == $stored_password ) {

    if ($UserInfoObj->UserHasAbility($auth_username, 'admin')) return true;
    if ($UserInfoObj->UserHasAbility($GuestUsername, $level)) return true;
    if ($UserInfoObj->UserHasAbility($auth_username, $level)) return true;

    if( $_SESSION['username'] != "" ) {
      PrintStandalonePage($LackProperAbilitiesFmt);
      exit;
    }
  } else {
    print "Bad username/password<BR>";
  }

  return false;
}

function CheckWikiPassword($page_passwd, $level, $page, $pagename) {
  global $GroupAttributesFmt,$DefaultPasswords, $AllowPassword;

  @$stored_passwd = $page["passwd$level"];
  if ($stored_passwd=="") {
    $grouppg = ReadPage(FmtPageName($GroupAttributesFmt,$pagename));
    @$stored_passwd = $grouppg["passwd$level"];
  }

  if (crypt($AllowPassword, $stored_passwd) == $stored_passwd) return true;
  if ($stored_passwd=="") { $stored_passwd = @$DefaultPasswords[$level]; }
  if ($stored_passwd=="") return true;
  
  foreach(array_merge($DefaultPasswords['admin'],$stored_passwd) as $pw)
    if (crypt($page_passwd, $pw)==$pw) return true;

  return false;
}

function PrintStandalonePage($prompt) {
  global $HTMLBodyFmt,$SessionAuthFmt, $AuthPageTitle;

  $action = $GLOBALS['action']; 
  PrintFmt($pagename, $HTMLStartFmt);
  PrintFmt($pagename, $PageStartFmt);
  PrintFmt($pagename, $AuthPageTitle);
  PrintFmt($pagename, $prompt);
  PrintFmt($pagename, $HTMLEndFmt);

}

function PrintEmbeddedPage($pagename, $prompt) {
  global $PageStartFmt, $PageEndFmt, $AuthPageTitle, $TimeFmt, $Now;

  $GLOBALS['LastModified'] = strftime($TimeFmt, $Now);
  $GLOBALS['LastModifiedBy'] = 'NoOne';
  $GLOBALS['LastModifiedHost'] = 'NoHost';
  $GLOBALS['HTMLTitle'] = $AuthPageTitle;
  $GLOBALS['GCount'] = 0;

  $AuthEmbeddedFmt = array( &$PageStartFmt,
			    $prompt,
			    &$PageEndFmt);
  $AuthEmbeddedFmt = FmtPageName($AuthEmbeddedFmt, $pagename);
  PrintFmt($pagename, $AuthEmbeddedFmt);
}

function GetUserLoginForm($no_title) {
  global $Author, $WikiTitle;

  if( $_SESSION['username'] == "" ) {

    if (!$no_title) {
      $login_form .= "<h1>Login to $WikiTitle</h1>";
    }

    $login_form .=
      "<form name='authform' action='{$_SERVER['REQUEST_URI']}?action=login' method='post'>
       <table>

       <tr>
           <td>Username:</td>
           <td><input name='username' value='$author' /></td>
       </tr>

       <tr>
           <td>Password:</td>
           <td><input name='user_pw' type='password' value='' /></td>
       </tr>
       <tr>
           <td align=left><input type='submit' value='Login' /></td>
           <td>&nbsp</td> 
       </tr>
       </table>
       </form>";

  } else {
    $login_form = 
      "Logged in as: " . $Author . "<br>
       <a href=\"" . $_SERVER['REQUEST_URI'] . "?action=logout\">Logout</a>";
    
  }
 
  return $login_form;
}

function GetWikiPasswordForm($pagename) {

  $password_form .=
    "<h1>Enter password for $pagename</h1>
     <form name='authform' action='{$_SERVER['REQUEST_URI']}' method='post'>
     <table>

     <tr>
         <td>Password:</td>
         <td><input type='password' name='wiki_pw' value='' /></td>
     </tr>
     <tr>
         <td align=left><input type='submit' value='Login' /></td>
         <td>&nbsp</td> 
     </tr>
     </table>
     </form>";

  return $password_form;
}

function GetLoginAction() {

  if( $_SESSION['username'] == "" ) {
    return "login";
  } else {
    return "logout";
  }

}

function HandleLogin($pagename) {
  global $LoginPage;

  if( $_SESSION['username'] != "" ) {
    if (isset($_SESSION['previous_page'])) {
      $prev_page = $_SESSION['previous_page'];
      unset($_SESSION['previous_page']);

      // Redirect to a URL if we have a referrer else go to
      // the page name we are handling login for
      if (preg_match("/^http/i", $prev_page)) {
	RedirectURL($prev_page);
      } else {
	Redirect($prev_page);
      }

    } else {
      Redirect($pagename);
    }
  } else {
    # Make sure both that a login page is defined and that we
    # are not currently at the login page. If no default abilities
    # are defined then the system would keep redirecting to the
    # login page since it is handled by PmWiki. Therefore if the
    # current page is login page when a login action is requested, we
    # just print out the backup standalone page.
    if (PageExists($LoginPage) && ($pagename != $LoginPage)) {
      // Remember the page that called for login, either by 
      // tracking the referer or by going to the page name
      // whom the login action was called on
      if (isset($_SERVER['HTTP_REFERER']) && $_SERVER['HTTP_REFERER'] != "") {
	$_SESSION['previous_page'] = $_SERVER['HTTP_REFERER'];
      } else {
	$_SESSION['previous_page'] = $pagename;
      }

      // Redirect to the login page to display username password form
      Redirect($LoginPage);
    } else {
      PrintStandalonePage( GetUserLoginForm(false) );
    }
  }
}

function HandleLogout($pagename) {
  $_SESSION['username'] = "";
  $_SESSION['user_pw']  = "";
  $_SESSION['wiki_pw']  = "";

  // Use referer for redirection if it is available
  if (isset($_SERVER['HTTP_REFERER']) && $_SERVER['HTTP_REFERER'] != "") {
    RedirectURL($_SERVER['HTTP_REFERER']);
  } else {
    Redirect($pagename);
  }
}

function RedirectURL($url) {
  global $RedirectDelay;
  clearstatcache();

  header("Location: $url");
  header("Content-type: text/html");
  print("<html><head>
    <meta http-equiv='Refresh' Content='$RedirectDelay; URL=$url'>
    <title>Redirect</title></head><body></body></html>");
  exit;
}

?>
